//**************************************//
//*  AES top module for SASEBO-AES     *//
//*    (c) AIST & Tohoku Univ.         *//
//*                                    *//
//*    File name   : SASEBO_AES_TOP.v  *//
//*    Version     : Version 1.01      *//
//*    Created     : APRIL/01/2007     *//
//*    Last update : OCTOBER/17/2007   *//
//**************************************//


/////////////////////////////
//         AES Top         //
/////////////////////////////
module AES_TOP(DinKey, Dout, Drdy, Krdy, ENC_DEC, RSTn, RSTn_HW, EN, CLK, BSY, Svld, Kvld, Dvld, Status);
  input  [127:0] DinKey;  // Data and key input
  output [127:0] Dout;    // Data output
  input  Drdy;            // Data input ready
  input  Krdy;            // Key input ready
  input  ENC_DEC;         // 0:Encryption 1:Decryption
  input  RSTn;            // Reset (Low active)
  input  RSTn_HW;         // Reset (Low active, Hardware Reset)
  input  EN;              // AES circuit enable
  input  CLK;             // System clock
  output BSY;             // Busy signal (even in busy, BSY goes 0 in one clock if new data is ready)
  output Svld;            // Status Register is Valid
  output Kvld;            // Key valid
  output Dvld;            // Data output valid
  output [3:0] Status;    // Status

  parameter [3:0] IDL = 0,
                  KSET_ENC = 1, KSET_DEC = 2,
                  TEST_ENC = 3, TEST_DEC = 4,
                  TEST_KSET_ENC =  7, TEST_KSET_DEC = 8,
                  TEST_LOOP_ENC =  9, TEST_LOOP_DEC = 10,
                  KEY_CHECK_ENC = 11, KEY_CHECK_DEC = 12,
                  ERROR = 15;
  reg  [3:0]   mode;
  reg  [3:0]   STrg;  // <bit3> 0: NORAMAL 1: ERROR
                      // <bit2> 1: Input key test failed
                      // <bit1> 1: Input key mismatch
                      // <bit0> 1: Power-up test failed
  reg  [127:0] Drg;   // Plaintext register for test
  reg  [127:0] Krg;   // Key register for test
  reg  [127:0] Tmprg; // Temporary register for test
  reg  ENC_DECrg, Krdyrg, Kvldrg, Drdyrg, Dvldrg, Dvldlast, DataRcved, ENC_DEClast;
  wire [127:0] Dinx, Keyx, Doutx;
  wire ENC_DECx, Krdyx, Drdyx, Kvldx, Dvldx;

  AES_EncDec AES_EncDec(Dinx, Keyx, Doutx, Drdyx, Krdyx, ENC_DECx, RSTn & RSTn_HW, EN, CLK, BSY, Kvldx, Dvldx);

  assign Status   = STrg;
  assign Keyx     = (mode == TEST_KSET_ENC || mode == TEST_KSET_DEC)? Krg: DinKey;
  assign Krdyx    = (mode == TEST_KSET_ENC || mode == TEST_KSET_DEC)? Krdyrg: Krdy;
  assign Dinx     = (mode == IDL)? DinKey: Drg;
  assign Drdyx    = (mode == IDL)? Drdy: Drdyrg;
  assign ENC_DECx = (mode == IDL || mode == KSET_ENC || mode == KSET_DEC)? ENC_DEC: ENC_DECrg;
  assign Dout     = Drg & {128{Dvldrg}};
  assign Dvld     = Dvldrg & (Dvldlast | ~Dvldx);
  assign Kvld     = Kvldrg;
  assign Svld     = (mode == IDL || mode == ERROR)? 1: 0;

  always @(posedge CLK) begin
    ENC_DEClast <= ENC_DEC;
    Dvldlast <= Dvldx;

    if ((RSTn & RSTn_HW) == 0) begin
      mode        <= TEST_KSET_ENC;
      STrg        <= 4'b1000;
      ENC_DECrg   <= 0;
      ENC_DEClast <= 0;
      Krdyrg      <= 1;
      Kvldrg      <= 0;
      Drdyrg      <= 0;
      Dvldrg      <= 0;
      Dvldlast    <= 0;
      DataRcved   <= 0;
      Krg         <= 128'h000102030405060708090a0b0c0d0e0f;
                             // Test Key from FIPS-197
      Drg         <= 128'h00112233445566778899aabbccddeeff;
                             // Test Plaintext from FIPS-197
    end
    else if (EN == 1) begin
      if (BSY == 0) begin

          case (mode)
            TEST_KSET_ENC: begin     // Encryption key set
              Krdyrg <= 0;
              mode   <= TEST_LOOP_ENC;
              Drdyrg <= 1;
            end

            TEST_LOOP_ENC: begin     // Loop encryption 128 times
              Drdyrg <= 0;
              if (Dvldx == 1) begin
                if (Krg[7:0] == 4'hf && Doutx != 128'h69c4e0d86a7b0430d8cdb78070b4c55a) begin
                  mode    <= ERROR;
                  STrg[0] <= 1;    // Power-up test failed
                end
                else begin
                  Tmprg     <= Drg;
                  Drg       <= Doutx;
                  mode      <= TEST_KSET_DEC;
                  ENC_DECrg <= 1;
                  Krdyrg    <= 1;
                end
              end
            end

            TEST_KSET_DEC: begin     // decryption key set
              Krdyrg <= 0;
              if (Krdyrg == 0 && Kvldx == 1) begin
                mode      <= TEST_LOOP_DEC;
                Drdyrg    <= 1;
              end
            end

            TEST_LOOP_DEC: begin     // Loop encryption 128 times
              Drdyrg <= 0;
              if (Dvldx == 1) begin
                if (Tmprg != Doutx) begin
                  mode    <= ERROR;
                  STrg[0] <= 1;   // Power-up test failed
                end
                ENC_DECrg <= 0;
                Krg       <= {Krg[0], Krg[127:1]};
                Krdyrg    <= 1;
                if (Krg[8:1] == 4'hf) begin
                   mode    <= IDL;
                   STrg[3] <= 0;
                end
                else
                   mode <= TEST_KSET_ENC;
              end
            end

            ERROR: begin
              STrg[3] <= 1;
            end

            IDL: begin
              if (ENC_DEC == 0 && Krdy == 1) begin
                mode      <= KSET_ENC;
                Dvldrg    <= 0;
                DataRcved <= 0;
              end
              else begin
                if (Drdy == 1 && Kvldx == 1)
                  DataRcved <= 1;
                else if (ENC_DEClast != ENC_DEC)
                  DataRcved <= 0;
                if (DataRcved == 1 && Dvldx == 1) begin
                  Drg    <= Doutx;
                  Dvldrg <= 1;
                end
              end
            end

            KSET_ENC: begin
              if (ENC_DEC == 1 && Krdy == 1) begin
                Kvldrg    <= 0;
                Tmprg     <= DinKey;
                mode      <= KSET_DEC;
              end
            end

            KSET_DEC: begin
              if (DinKey != Tmprg) begin
                mode <= ERROR;
                STrg <= 4'b1010;  // Input key mismatch
              end
              else if (Kvldx == 1) begin
                mode      <= TEST_ENC;
                STrg      <= 4'b0000; 
                ENC_DECrg <= 0;
                Drg       <= 128'h00000000000000000000000000000000;
                Drdyrg    <= 1;
              end
            end

            TEST_ENC: begin     // Encrypt 0 data with encryption key
              Drdyrg <= 0;
              if (Dvldx == 1) begin
                mode      <= TEST_DEC;
                ENC_DECrg <= 1;
                Drg       <= Doutx;
                Drdyrg    <= 1;
              end
            end

            TEST_DEC: begin     // Decrypt cipher text with decryption key
              Drdyrg <= 0;
              if (Drdyrg == 0 && Dvldx == 1) begin
                if (Doutx != 128'h00000000000000000000000000000000) begin
                  mode    <= ERROR;  // Input key test failed
                  STrg[2] <= 1;
                end
                else begin
                  Kvldrg  <= Kvldx;
                  mode    <= IDL;
                end
              end
            end
          endcase
      end
    end
  end
endmodule 


///////////////////////////////////////
// AES for Encryption and Decryption //
///////////////////////////////////////

module AES_EncDec(Din, Key, Dout, Drdy, Krdy, ENC_DEC, RSTn, EN, CLK, BSY, Kvld, Dvld);
  input  [127:0] Din;  // Data input
  input  [127:0] Key;  // Key input
  output [127:0] Dout; // Data output
  input  Drdy;         // Data input ready
  input  Krdy;         // Key input ready
  input  ENC_DEC;      // 0:Encryption 1:Decryption
  input  RSTn;         // Reset (Low active)
  input  EN;           // AES circuit enable
  input  CLK;          // System clock
  output BSY;          // Busy signal
  output Dvld;         // Data output valid
  output Kvld;         // Data output valid

wire EN_E, EN_D;
wire [127:0] Dout_E, Dout_D;
wire BSY_E, BSY_D;
wire Dvld_E, Dvld_D;
wire Kvld_E, Kvld_D;

assign EN_E = (~ENC_DEC) & EN;
assign EN_D =  ENC_DEC & EN;
assign BSY  = BSY_E | BSY_D;
assign Dvld = Dvld_E & (~ENC_DEC) | Dvld_D & ENC_DEC;
assign Kvld = Kvld_E & Kvld_D;
assign Dout = (ENC_DEC == 0)? Dout_E: Dout_D;

AES_ENC AES_ENC(Din, Key, Dout_E, Drdy, Krdy, RSTn, EN_E, CLK, BSY_E, Kvld_E, Dvld_E);
AES_DEC AES_DEC(Din, Key, Dout_D, Drdy, Krdy, RSTn, EN_D, CLK, BSY_D, Kvld_D, Dvld_D);

endmodule